import numpy as np
from pyPhasesML.datapipes.DataPipe import DataPipe


class RecordMap(DataPipe):
    def __init__(self, dataset: DataPipe, mapping, mappingLengths=None):
        self.dataset = dataset

        if mappingLengths is not None:
            mapping = self.convertMappingToSegments(mapping, mappingLengths)

        self.mapping = mapping
        self.currentIndex = 0

    def convertMappingToSegments(self, recordMapping, segmentLengths):
        segmentLengthsCum = np.cumsum([0] + segmentLengths)
        segments = np.arange(sum(segmentLengths))
        segments = [segments[segmentLengthsCum[recordMapping[i]] : segmentLengthsCum[recordMapping[i] + 1]] for i in range(len(recordMapping))]

        return [item for sublist in segments for item in sublist]

    def __len__(self):
        return len(self.mapping)

    def __getitem__(self, index):
        return self.dataset[self.mapping[index]]



import gc

import numpy as np
from pyPhasesML.scorer.ScorerTorch import ScorerTorch
import torch
from tqdm import tqdm

from pyPhasesML.adapter.torch.CSVLogger import CSVLogger
from pyPhasesML.adapter.torch.LoadOptimizer import LoadOptimizer
from pyPhasesML.adapter.torch.SystemCheckPoint import SystemCheckPoint
from pyPhasesML.adapter.torch.CheckPoint import CheckPoint

from ..DataSet import DataSet, TrainingSetLoader
from ..ModelAdapter import ModelAdapter


class ModelTorchAdapter(ModelAdapter):
    model: torch.nn.Module
    useGPU = torch.cuda.is_available()

    def __init__(self, options=None) -> None:
        super().__init__(options)
        self.cbs = []
        self.skipValidation = False
        self.metricDefinitions = {}
        self.batchScheduler = None
        self.optimizer = None

    def prepareX(self, x, validation=False):
        x = x.permute(0, 2, 1)
        return x

    def prepareDataAdapter(self, datasetOrTuple, validation=False):
        x, y = datasetOrTuple

        if isinstance(x, list):
            x = torch.stack(x, axis=0)
        if isinstance(y, list):
            y = torch.stack(y, axis=0)

        if not torch.is_tensor(x):
            x = torch.tensor(x)
        if not torch.is_tensor(y):
            y = torch.tensor(y)

        dataset = DataSet(x, y)

        if self.useGPU:
            dataset.x = dataset.x.cuda()
            dataset.y = dataset.y.cuda()

        return self.prepareData(dataset, validation)

    def registerCB(self, cb):
        self.cbs.append(cb)
        self.cbs.sort(key=lambda x: x.priority)
        cb.trigger("register", self)

    def initAdapter(self, findingLearningRate=False):
        self.batchScheduler = None
        config = self.config

        self.registerCB(LoadOptimizer(config))
        self.registerCB(CSVLogger(config))
        self.registerCB(SystemCheckPoint(config))
        self.registerCB(CheckPoint(config))

    def trigger(self, eventName, *args, **kwargs):
        for cb in self.cbs:
            cb.trigger(eventName, *args, **kwargs)

        super().trigger(eventName, *args, **kwargs)

    def prepareForScore(self, targets, prediction):
        lastDimension = self.config.numClasses if self.oneHotDecoded else 1
        targets = targets.reshape(-1, lastDimension)
        prediction = prediction.reshape(-1, lastDimension)
        return targets, prediction

    def scoreBatch(self, scorer, truth, predictions, processList=None):
        results = scorer.score(truth, predictions, trace=True)

        if processList is not None:
            processList.set_postfix({m: results[m] for m in scorer.metrics})

    def scoreResult(self, scorer):
        results = scorer.scoreAllRecords()
        metrics = {m: results[m] for m in scorer.metrics}
        justPrint = []

        if "confusion" in scorer.results:
            justPrint.append(scorer.results["confusion"])

        return metrics, justPrint

    def getScorer(self):
        scorer = ScorerTorch(self.config.numClasses, trace=True)
        scorer.metrics = self.validationMetrics
        scorer.trace = True
        scorer.ignoreClasses = (
            [self.ignoreClassIndex] if self.ignoreClassIndex is not None else []
        )
        return scorer

    def validate(self, validationData):
        self.trigger("validationStart", self, validationData)
        model = self.model
        model.eval()
        scorer = self.getScorer()

        batchCount = len(validationData)
        processList = tqdm(range(batchCount), disable=(not self.showProgress))
        processList.set_description("Validation")
        batchGenerator = iter(validationData)

        for batchIndex in processList:
            validationBatch = batchGenerator.__next__()
            x, y = self.prepareDataAdapter(validationBatch, validation=True)

            # Run model
            with torch.no_grad():
                output = model(x)
            batchPredictions = self.mapOutputForPrediction(output)

            if len(self.validationMetrics) > 0:
                y, batchPredictions = self.prepareForScore(y, batchPredictions)

                self.scoreBatch(scorer, y, batchPredictions, processList)

            del batchPredictions
            del x
            del y
            del output
            gc.collect()

        if len(self.validationMetrics) > 0:
            metrics, justPrintMetrics = self.scoreResult(scorer)
            self.trigger("validationEnd", self, metrics, scorer)
            return metrics, justPrintMetrics

        self.trigger("validationEnd", self, {}, scorer)
        return None, None

    def prepareTargetsForLoss(self, targets, oneHotDecoded=None):
        oneHotDecoded = self.oneHotDecoded if oneHotDecoded is None else oneHotDecoded
        mask = None
        if oneHotDecoded:
            targets = targets.reshape(-1, self.config.numClasses)
            if self.ignoreClassIndex is not None:
                mask = targets.eq(0).sum(-1) < self.config.numClasses
                targets = targets[mask]
        else:
            targets = targets.reshape(-1)

            if self.ignoreClassIndex is not None:
                mask = targets != self.ignoreClassIndex
                targets = targets[mask]

        return targets, mask

    def train(self, dataset: TrainingSetLoader):
        self.bestModelPath = None
        self.finish = False

        self.trigger("trainingStart", self, dataset)

        lossCriterion = self.getLossFunction()
        self.epoch = self.startEpoch

        while self.maxEpochs is None or self.epoch < self.maxEpochs:
            # Put in train mode

            self.model.train(True)
            self.runningStats = {"loss": 0.0}
            batchesPerEpoch = len(dataset.trainingData)
            processList = tqdm(range(batchesPerEpoch), disable=(not self.showProgress))
            processList.set_description(f"EPOCH {self.epoch}")
            batchGenerator = iter(dataset.trainingData)

            for batchIndex in processList:
                trainBatch = next(batchGenerator)
                batchFeats, targs = self.prepareDataAdapter(trainBatch)
                targs, mask = self.prepareTargetsForLoss(targs)

                self.optimizer.zero_grad()

                # check if there are any targets left
                if len(targs) == 0:
                    continue

                # for batchFeat in batchFeats:
                output = self.model(batchFeats)
                output = self.mapOutputForLoss(output, mask)
                loss = lossCriterion(output, targs)
                ownStats = hasattr(lossCriterion, "stats")

                if ownStats:
                    processList.set_postfix(ordered_dict=lossCriterion.stats)
                    for stat, value in lossCriterion.stats.items():
                        if stat not in self.runningStats:
                            self.runningStats[stat] = value
                        else:
                            self.runningStats[stat] += value

                # Backpropagation
                loss.backward()

                self.optimizer.step()

                # Perform one optimization step
                currentBatchLoss = loss.data.cpu().numpy()
                if np.isnan(currentBatchLoss):
                    self.model(batchFeats)
                    lossCriterion(output, targs)
                    raise Exception("batch loss should be a number")

                if self.batchScheduler is not None:
                    self.batchScheduler.step()

                self.runningStats["loss"] += currentBatchLoss
                self.runningStats["lr"] = self.optimizer.param_groups[0]["lr"]

                self.trigger("batchEnd", self, batchIndex)

                del output
                del targs
                del loss
                gc.collect()
                currentCount = processList.n + 1
                processList.set_postfix(
                    ordered_dict={
                        n: v / currentCount for n, v in self.runningStats.items()
                    }
                )

            self.runningStats = {
                n: v / batchesPerEpoch for n, v in self.runningStats.items()
            }

            self.epoch += 1

            if not self.skipValidation:
                self.validate(dataset.validationData)

            if self.finish:
                break

        self.trigger("trainingEnd", self)

        self.fullEpochs = self.epoch
        self.trigger("trainEnd", self)
        if self.bestModelPath is not None:
            self.log("load best Model: %s" % self.bestModelPath)
            self.loadState(self.load(self.bestModelPath))
        return self.model

    def getModelPath(self):
        return self.logPath

    def build(self):
        torchSeed = 2
        torch.manual_seed(torchSeed)

        if self.useGPU:
            torch.cuda.manual_seed(torchSeed)
            self.model.cuda()

    def summary(self):
        pytorch_total_params = sum(
            p.numel() for p in self.model.parameters() if p.requires_grad
        )
        self.log("Total trainable Parameters: %i" % (pytorch_total_params))
        self.parameter = pytorch_total_params

        return str(self.model)

    def cleanUp(self):
        if self.useGPU:
            torch.cuda.empty_cache()

    def save(self, path):
        torch.save(self.model.state_dict(), path)

    def load(self, path):
        return torch.load(
            path,
            map_location=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
        )

    def loadState(self, state):
        if isinstance(state, torch.nn.Module):
            state = state.state_dict()
        return self.model.load_state_dict(state)

    def mapOutputForLoss(self, output, mask=None):
        output = (
            output.reshape(-1, self.config.numClasses)
            if self.oneHotDecoded
            else output.flatten()
        )
        return output[mask] if mask is not None else output

    def predict(self, input, get_likelihood=False, returnNumpy=True):
        with torch.no_grad():
            if not torch.is_tensor(input):
                input = torch.tensor(input)

            batchSize, _, _ = input.shape

            if self.useGPU:
                input = input.cuda()

            input = self.prepareX(input)
            model = self.getModelEval()
            out = model(input)

            predictions = self.mapOutputForPrediction(out)

            if self.config.numClasses > 0 and self.oneHotDecoded:
                predictions = predictions.reshape(batchSize, -1, self.config.numClasses)

                if not get_likelihood:
                    predictions = torch.argmax(predictions, dim=2)

            if returnNumpy:
                predictions = predictions.detach().cpu().numpy()

            return predictions



import math
import numpy as np
from pyPhases import classLogger


class NotUniqueException(Exception):
    pass


class NotCompleteException(Exception):
    pass

class SplitsMisconfiguredException(Exception):
    pass


@classLogger
class DataversionManager:
    def __init__(self, groupedRecords, splits=None, seed=None) -> None:
        splits = {} if splits is None else splits

        self.splits = self.normalizeSlicesDict(splits)
        self.seed = seed
        self.removedRecords = {}
        self.splitHasSubs = []

        self.remainingSplit = None

        # shuffle the records
        groups = list(groupedRecords.keys())
        if seed is not None:
            np.random.seed(seed)
            np.random.shuffle(groups)
            groupedRecords = {g: groupedRecords[g] for g in groups}

        self.groupedRecords = groupedRecords

    def normalizeSlices(self, slices):
        slices = slices if isinstance(slices, list) else [slices]
        splitSlices = []

        for splitSlice in slices:
            splitSlice = (
                slice(*[int(index) for index in splitSlice.split(":")])
                if isinstance(splitSlice, str)
                else splitSlice
            )
            splitSlices.append(splitSlice)
        return splitSlices

    def normalizeSlicesDict(self, slices):
        return {k: self.normalizeSlices(v) for k, v in slices.items()}

    def addSplitBySlices(self, name, slices):
        self.splits[name] = self.normalizeSlices(slices)

    def addSplitsByFold(
        self,
        mainSplit: str,
        foldSplit: str,
        recordSlice: slice | str,
        foldCount: int,
        currentFold: int,
    ):
        if currentFold >= foldCount:
            raise SplitsMisconfiguredException(
                "Fold %i is higher than fold count %i" % (currentFold, foldCount)
            )

        subSlice, mainSlice = self.getSplitsByPercentage(
            recordSlice, subLength=1 / foldCount, subPosition=currentFold
        )

        self.addSplitBySlices(foldSplit, subSlice)
        self.addSplitBySlices(mainSplit, mainSlice)

    def getSplitsByPercentage(
        self, recordSlice: slice | str, subLength, subPosition=0, ceil=True
    ):
        recordSlice = self.normalizeSlices(recordSlice)[0]
        recordSliceLength = recordSlice.stop - recordSlice.start
        subLength = (
            math.ceil(subLength * recordSliceLength)
            if ceil
            else math.floor(subLength * recordSliceLength)
        )
        subSlice = slice(
            recordSlice.start + int(subLength * subPosition),
            recordSlice.start + int(subLength * (subPosition + 1)),
        )

        if subPosition == 0:
            mainSlice = slice(subSlice.stop, recordSlice.stop)
        else:
            mainSlice = [
                slice(recordSlice.start, subSlice.start),
                slice(subSlice.stop, recordSlice.stop),
            ]

        return subSlice, mainSlice

    def getRemainingSplit(self):
        if len(self.splits) == 0:
            return slice(0, len(self.groupedRecords))
        elif len(self.splits) == 1:
            curSplit = list(self.splits.values())[0]
            if curSplit.start == 0:
                return slice(curSplit.stop, len(self.groupedRecords))
            elif curSplit.stop == len(self.groupedRecords):
                return slice(0, curSplit.start)
            else:
                raise SplitsMisconfiguredException(
                    "Current split is not at the beginning or end of the dataset"
                )
        else:
            raise SplitsMisconfiguredException(
                "Remaining splits can only be used if there is only one split."
            )

    def addSplitByRemaining(self, name: str, percentage: float, ceil=True):
        remainingSplit = (
            self.getRemainingSplit()
            if self.remainingSplit is None
            else self.remainingSplit
        )
        recordCount = remainingSplit.stop - remainingSplit.start
        length = (
            math.ceil(recordCount * percentage)
            if ceil
            else math.floor(recordCount * percentage)
        )
        split = slice(remainingSplit.start, remainingSplit.start + length)

        self.remainingSplit = slice(split.stop, remainingSplit.stop)

        self.addSplitBySlices(name, split)

    def groupDatasetBySplit(
        self, datasetName, splits, groupedRecords, removedRecords=None
    ):
        split = splits[datasetName]
        # check if split is a list of recordids
        if isinstance(split, dict):
            mainSplit = split["mainsplit"]
            mainRecordIds = self.getRecordsForSplit(mainSplit)
            recordIds = [mainRecordIds[i] for i in split["indexes"]]
        else:
            splits = self.normalizeSlicesDict(splits)
            slicesForDataset = (
                splits[datasetName]
                if isinstance(splits[datasetName], list)
                else self.normalizeSlices(splits[datasetName])
            )
            # get record ids for splits and flatten
            groupIds = [
                recordId
                for s in slicesForDataset
                for recordId in list(groupedRecords.keys())[s]
            ]
            recordIds = [
                record for group in groupIds for record in groupedRecords[group]
            ]

        if bool(removedRecords):
            recordIds = [
                record
                for index, record in enumerate(recordIds)
                if index not in removedRecords
            ]

        return recordIds

    def groupDatasetsBySplit(self, datasetNames, splits, groupedRecords):
        recordSlices = {}
        for datasetName in datasetNames:
            recordSlices[datasetName] = self.groupDatasetBySplit(
                datasetName, splits, groupedRecords
            )

        return recordSlices

    def getRecordsForSplit(self, datasetName):
        removedRecords = self.removedRecords.get(datasetName, None)
        return self.groupDatasetBySplit(
            datasetName, self.splits, self.groupedRecords, removedRecords
        )

    def getRecordIndexesForSplit(self, datasetName):
        removedRecords = self.removedRecords.get(datasetName, None)
        recordIds = self.groupDatasetBySplit(
            datasetName, self.splits, self.groupedRecords, removedRecords
        )
        return [i for i, r in enumerate(self.groupedRecords) if r in recordIds]

    def validatDatasetVersion(self, raiseException=True):
        allDBRecordIds = self.groupedRecords
        # remove splits keys that have subsplits
        splits = {k: v for k, v in self.splits.items() if k not in self.splitHasSubs}
        datasetNames = list(splits.keys())

        recordSlices = self.groupDatasetsBySplit(datasetNames, splits, allDBRecordIds)

        flattenAllRecords = [r for elem in allDBRecordIds.values() for r in elem]
        flattenUsedRecords = [r for elem in recordSlices.values() for r in elem]

        # check if all records are unique and present in the dataset splits
        complete = len(flattenAllRecords) == len(flattenUsedRecords)
        unique = len(flattenUsedRecords) == len(set(flattenUsedRecords))
        if not unique:
            error = "There are duplicate records in the dataset splits "
            self.logError(error)
            if raiseException:
                raise NotUniqueException(error)

        if not complete:
            error = (
                "Not all records are used (%i records are missing), overall %s groups"
                % (
                    len(flattenAllRecords) - len(flattenUsedRecords),
                    len(allDBRecordIds.keys()),
                )
            )
            self.logError(error)

            if raiseException:
                raise NotCompleteException(error)

        if complete and unique:
            self.logSuccess("All records are unique and present in the dataset splits")

    def removeRecordIndexesFromSplit(self, datasetName, recordIndexes):
        self.removedRecords[datasetName] = recordIndexes

    def addVirtualSplitByMainSplitIndexes(self, mainsplit, datasetName, recordIndexes):
        self.splitHasSubs.append(mainsplit)
        self.splits[datasetName] = {
            "mainsplit": mainsplit,
            "indexes": list(recordIndexes),
        }



from pyPhasesML.datapipes.DataLoader import DataPipe
from .DataManipulation import DataManipulation
from .DataSet import DataSet, TrainingSetLoader
from .FeatureExtraction import FeatureExtraction
from .Model import Model
from .ModelManager import ModelManager
from .scorer.Scorer import Scorer
from .DataversionManager import DataversionManager

from .datapipes import *



import numpy as np

from typing import List
from pyPhasesML.datapipes import RecordMap
from pyPhasesML.datapipes.DataPipe import DataPipe


class MultiSourceMap(DataPipe):
    def __init__(self, dataset: List[RecordMap] = None, seed=2024):
        dataset = dataset or []
        
        self.datapipe = []
        self.length = 0
        self.recordMapIndex = np.empty((0, 2), dtype=int)
        self.seed = seed

        for recordmap in dataset:
            self.addRecordMap(recordmap)



    def addRecordMap(self, recordmap: RecordMap):
        datasetIndex = len(self.datapipe)
        self.datapipe.append(recordmap)
        self.length += len(recordmap)

        recordMapping = np.empty((len(recordmap), 2), dtype=int)
        recordMapping[:, 0] = datasetIndex
        recordMapping[:, 1] = np.arange(0, len(recordmap), dtype=int)

        self.recordMapIndex = np.append(self.recordMapIndex, recordMapping, axis=0)

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        datasetIndex, recordIndex = self.recordMapIndex[index]
        return self.datapipe[datasetIndex][recordIndex]
    
    def shuffle(self):
        np.random.seed(self.seed)
        np.random.shuffle(self.recordMapIndex)
